from copy import deepcopy
import numpy as np
import blosc2
import math
from typing import Dict, Optional, Union, List, Tuple
from pathlib import Path
import os
from med_blosc2.meta import Meta, MetaSpatial, MetaBlosc2
from med_blosc2.utils import is_serializable

MED_BLOSC2_SUFFIX = "mb2nd"
MED_BLOSC2_VERSION = "v0"
MED_BLOSC2_DEFAULT_PATCH_SIZE = 192


class MedBlosc2:
    def __init__(self,
                 array: Union[np.ndarray, str, Path],
                 spacing: Optional[Union[List, Tuple, np.ndarray]] = None,
                 origin: Optional[Union[List, Tuple, np.ndarray]] = None,
                 direction: Optional[Union[List, Tuple, np.ndarray]] = None,
                 meta: Optional[Union[Dict, Meta]] = None,
                 mmap: bool = False,
                 num_threads: int = 1,
                 mode: str = 'r',
                 copy: Optional['MedBlosc2'] = None) -> None:
        """Initializes a MedBlosc2 instance.

        The MedBlosc2 file format (".mb2nd") is a Blosc2-compressed container
        with standardized metadata support for N-dimensional medical images.

        Args:
            array (Union[np.ndarray, str, Path]): Input data or file path. Use
                a numpy ndarray for in-memory arrays. Use a string or Path to
                load a ".b2nd" or ".mb2nd" file.
            spacing (Optional[Union[List, Tuple, np.ndarray]]): Spacing per
                axis. Provide a list/tuple/ndarray with length equal to the
                number of dimensions (e.g., [sx, sy, sz]).
            origin (Optional[Union[List, Tuple, np.ndarray]]): Origin per axis.
                Provide a list/tuple/ndarray with length equal to the number of
                dimensions.
            direction (Optional[Union[List, Tuple, np.ndarray]]): Direction
                cosine matrix. Provide a 2D list/tuple/ndarray with shape
                (ndims, ndims).
            meta (Optional[Dict | Meta]): Free-form metadata dictionary or Meta
                instance. Must be JSON-serializable when saving. 
                If meta is passed as a Dict, it will internally be converted into a Meta object with the dict being interpreted as meta.image metadata.
            mmap (bool): Whether to keep the loaded array memory-mapped when
                loading from disk. If true, MedBlosc2.array will be an blosc2.ndarray.NDArray, else np.ndarray.
            num_threads (int): Number of threads for Blosc2 operations.
            mode (str): Blosc2 open mode
                - 'r': read-only, must exist (Default)
                - 'a': read/write, create if doesn't exist (Currently not supported)
                - 'w': create, overwrite if it exists (Currently not supported)
            copy (Optional[MedBlosc2]): Another MedBlosc2 instance to copy
                metadata fields from.
        """
        if isinstance(array, (str, Path)) and (spacing is not None or origin is not None or direction is not None or meta is not None or copy is not None):
            raise RuntimeError("Spacing, origin, direction, meta or copy cannot be set if array is a string to load an image.")
        
        if mode != 'r':
            raise NotImplementedError("Currently 'r' is the only supported blosc2 open mode.")
        
        if isinstance(array, (str, Path)):
            array, meta = self._load(array, num_threads, mode, mmap)
            spacing = meta.spatial.spacing
            origin = meta.spatial.origin
            direction = meta.spatial.direction

        self.array = array

        # Validate meta: Must be None, a dictionary or a Meta object. Convert to Meta object if necessary.
        if meta is not None:
            if not isinstance(meta, (dict, Meta)):
                raise ValueError("Meta must be None, a dict or a Meta object.")
            if isinstance(meta, dict):
                meta = Meta(image=meta)
        else:
            meta = Meta()
        self.meta = meta
        self.meta._med_blosc2_version = MED_BLOSC2_VERSION
        meta_spatial = MetaSpatial(spacing, origin, direction)
        meta_spatial._validate_and_cast(self.ndims)
        self.meta.spatial = meta_spatial
        
        # If copy is set, copy fields from the other Nifti instance
        if copy is not None:
            self.meta.copy_from(copy.meta)

    @property
    def spacing(self):
        """Returns the image spacing.

        Returns:
            list: The image spacing with length equal to the number of
            dimensions.
        """
        return self.meta.spatial.spacing
    
    @property
    def origin(self):
        """Returns the image origin.

        Returns:
            list: The image origin with length equal to the number of
            dimensions.
        """
        return self.meta.spatial.origin
    
    @property
    def direction(self):
        """Returns the image direction.

        Returns:
            list: The image direction with shape (ndims, ndims).
        """
        return self.meta.spatial.direction

    @property
    def affine(self) -> np.ndarray:
        """Computes the affine transformation matrix for the image.

        Returns:
            list: Affine matrix with shape (ndims + 1, ndims + 1).
        """
        spacing  = np.array(self.spacing) if self.spacing is not None else np.ones(self.ndims)
        origin  = np.array(self.origin) if self.origin is not None else np.zeros(self.ndims)
        direction = np.array(self.direction) if self.direction is not None else np.eye(self.ndims)
        affine = np.eye(self.ndims+1)
        affine[:self.ndims, :self.ndims] = direction @ np.diag(spacing)
        affine[:self.ndims, self.ndims] = origin
        return affine.tolist()
    
    @property
    def translation(self):
        """Extracts the translation vector from the affine matrix.

        Returns:
            list: Translation vector with length equal to the number of
            dimensions.
        """
        return np.array(self.affine)[:-1, -1].tolist()

    @property
    def scale(self):
        """Extracts the scaling factors from the affine matrix.

        Returns:
            list: Scaling factors per axis with length equal to the number of
            dimensions.
        """
        scales = np.linalg.norm(np.array(self.affine)[:-1, :-1], axis=0)
        return scales.tolist()

    @property
    def rotation(self):
        """Extracts the rotation matrix from the affine matrix.

        Returns:
            list: Rotation matrix with shape (ndims, ndims).
        """
        rotation_matrix = np.array(self.affine)[:-1, :-1] / np.array(self.scale)
        return rotation_matrix.tolist()

    @property
    def shear(self):
        """Computes the shear matrix from the affine matrix.

        Returns:
            list: Shear matrix with shape (ndims, ndims).
        """
        scales = np.array(self.scale)
        rotation_matrix = np.array(self.rotation)
        shearing_matrix = np.dot(rotation_matrix.T, np.array(self.affine)[:-1, :-1]) / scales[:, None]
        return shearing_matrix.tolist()
    
    @property
    def shape(self):
        """Returns the shape of the array.

        Returns:
            tuple: Shape of the underlying array.
        """
        return self.array.shape
    
    @property
    def ndims(self) -> int:
        """Returns the number of dimensions of the image.

        Returns:
            int: Number of dimensions of the image (2D, 3D, or 4D).
        """
        return len(self.array.shape)

    def _load(
            self,
            filepath: Union[str, Path], 
            num_threads: int = 1,
            mode: str = 'r',
            mmap: bool = False
        ):
        """Loads a Blosc2-compressed file. Both MedBlosc2 ('.mb2nd') and Blosc2 ('.b2nd') files are supported.

        WARNING:
            MedBlosc2 supports both ".b2nd" and ".mb2nd" files. The MedBlosc2
            format standard and standardized metadata are honored only for
            ".mb2nd". For ".b2nd", metadata is ignored when loading.

        Args:
            filepath (Union[str, Path]): Path to the Blosc2 file to be loaded.
                The filepath needs to have the extension ".b2nd" or ".mb2nd".
            num_threads (int): Number of threads to use for loading the file.
            mode (str): Blosc2 open mode (e.g., "r", "a").
            mmap (bool): Whether to keep the array memory-mapped.

        Returns:
            Tuple[blosc2.ndarray, dict]: Loaded data and its metadata.

        Raises:
            RuntimeError: If the file extension is not ".b2nd" or ".mb2nd".
        """
        if not str(filepath).endswith(".b2nd") and not str(filepath).endswith(f".{MED_BLOSC2_SUFFIX}"):
            raise RuntimeError(f"MedBlosc2 requires '.b2nd' or '.{MED_BLOSC2_SUFFIX}' as extension.")
        blosc2.set_nthreads(num_threads)
        dparams = {'nthreads': num_threads}
        array = blosc2.open(urlpath=str(filepath), dparams=dparams, mode=mode, mmap_mode=mode)
        blosc2_metadata = dict(array.schunk.meta)
        meta = Meta()
        if str(filepath).endswith(f".{MED_BLOSC2_SUFFIX}"):
            if "med_blosc2" not in blosc2_metadata:
                raise RuntimeError(f"The header of the .{MED_BLOSC2_SUFFIX} is missing the 'med_blosc2' attribute.")
            meta = Meta.from_dict(blosc2_metadata["med_blosc2"])
        if not mmap:
            array = array[...]
        return array, meta

    def save(
            self,
            filepath: Union[str, Path],
            patch_size: Optional[Union[int, List, Tuple]] = 'default',  # 'default' means that the default of 192 is used. However, if set to 'default', the patch_size will be skipped if self.patch_size is set from a previously loaded MedBlosc2 image. In that case the self.patch_size is used.
            chunk_size: Optional[Union[int, List, Tuple]]= None,
            block_size: Optional[Union[int, List, Tuple]] = None,
            clevel: int = 8,
            codec: blosc2.Codec = blosc2.Codec.ZSTD,
            num_threads: int = 1
        ):
        """Saves the array to a Blosc2-compressed file. Both MedBlosc2 ('.mb2nd') and Blosc2 ('.b2nd') files are supported.

        WARNING:
            MedBlosc2 supports both ".b2nd" and ".mb2nd" files. The MedBlosc2
            format standard and standardized metadata are honored only for
            ".mb2nd". For ".b2nd", metadata is ignored when saving.

        Args:
            filepath (Union[str, Path]): Path to save the file. Must end with
                ".b2nd" or ".mb2nd".
            patch_size (Optional[Union[int, List, Tuple]]): Patch size hint for
                chunk/block optimization. Provide an int for isotropic sizes or
                a list/tuple with length equal to the number of dimensions.
                Use "default" to use the default patch size of 192.
            chunk_size (Optional[Union[int, List, Tuple]]): Explicit chunk size.
                Provide an int or a tuple/list with length equal to the number
                of dimensions, or None to let Blosc2 decide. Ignored when
                patch_size is not None.
            block_size (Optional[Union[int, List, Tuple]]): Explicit block size.
                Provide an int or a tuple/list with length equal to the number
                of dimensions, or None to let Blosc2 decide. Ignored when
                patch_size is not None.
            clevel (int): Compression level from 0 (no compression) to 9
                (maximum compression).
            codec (blosc2.Codec): Compression codec to use.
            num_threads (int): Number of threads to use for saving the file.

        Raises:
            RuntimeError: If the file extension is not ".b2nd" or ".mb2nd".
        """
        if not str(filepath).endswith(".b2nd") and not str(filepath).endswith(f".{MED_BLOSC2_SUFFIX}"):
            raise RuntimeError(f"MedBlosc2 requires '.b2nd' or '.{MED_BLOSC2_SUFFIX}' as extension.")
        blosc2.set_nthreads(num_threads)
        dparams = {'nthreads': num_threads}

        if patch_size is not None and patch_size != "default" and (self.ndims == 1 or self.ndims > 3):
            raise NotImplementedError("Chunk and block size optimization based on patch size is only implemented for 2D and 3D images. Please set the chunk and block size manually or set to None for blosc2 to determine a chunk and block size.")
        if patch_size is not None and patch_size != "default" and (chunk_size is not None or block_size is not None):
            raise RuntimeError("patch_size and chunk_size / block_size cannot both be explicitly set.")

        if patch_size == "default": 
            if self.meta._blosc2.patch_size is not None:  # Use previously loaded patch size, when patch size is not explicitly set and a patch size from a previously loaded image exists
                patch_size = self.meta._blosc2.patch_size
            else:  # Use default patch size, when patch size is not explicitly set and no patch size from a previously loaded image exists
                patch_size = [MED_BLOSC2_DEFAULT_PATCH_SIZE] * self.ndims

        patch_size = [patch_size] * self.ndims if isinstance(patch_size, int) else patch_size

        if patch_size is not None:
            chunk_size, block_size = self.comp_blosc2_params(self.array.shape, patch_size)

        meta_blosc2 = MetaBlosc2(chunk_size, block_size, patch_size)
        meta_blosc2._validate_and_cast(self.ndims)
        self.meta._blosc2 = meta_blosc2
        
        metadata = None
        if str(filepath).endswith(f".{MED_BLOSC2_SUFFIX}"):
            metadata = {"med_blosc2": self.meta.to_dict()}

        if not is_serializable(metadata):
            raise RuntimeError("Metadata is not serializable.")

        cparams = {'codec': codec, 'clevel': clevel,}
        self.array = np.ascontiguousarray(self.array[...])  # Needs to overwrite self.array to ensure there is no opened blosc2 overwriting itself
        if Path(filepath).is_file():
            os.remove(str(filepath))
        blosc2.asarray(self.array, urlpath=str(filepath), chunks=self.meta._blosc2.chunk_size, blocks=self.meta._blosc2.block_size, cparams=cparams, dparams=dparams, mmap_mode='w+', meta=metadata)

    def comp_blosc2_params(
            self,
            image_size: Tuple[int, int, int, int],
            patch_size: Union[Tuple[int, int], Tuple[int, int, int]],
            bytes_per_pixel: int = 4,  # 4 byte are float32
            l1_cache_size_per_core_in_bytes: int = 32768,  # 1 Kibibyte (KiB) = 2^10 Byte;  32 KiB = 32768 Byte
            l3_cache_size_per_core_in_bytes: int = 1441792, # 1 Mibibyte (MiB) = 2^20 Byte = 1.048.576 Byte; 1.375MiB = 1441792 Byte
            safety_factor: float = 0.8  # we dont will the caches to the brim. 0.8 means we target 80% of the caches
        ):
        """
        Computes a recommended block and chunk size for saving arrays with Blosc v2.

        Blosc2 NDIM documentation:
        "Having a second partition allows for greater flexibility in fitting different partitions to different CPU cache levels. 
        Typically, the first partition (also known as chunks) should be sized to fit within the L3 cache, 
        while the second partition (also known as blocks) should be sized to fit within the L2 or L1 caches, 
        depending on whether the priority is compression ratio or speed." 
        (Source: https://www.blosc.org/posts/blosc2-ndim-intro/)

        Our approach is not fully optimized for this yet. 
        Currently, we aim to fit the uncompressed block within the L1 cache, accepting that it might occasionally spill over into L2, which we consider acceptable.

        Note: This configuration is specifically optimized for nnU-Net data loading, where each read operation is performed by a single core, so multi-threading is not an option.

        The default cache values are based on an older Intel 4110 CPU with 32KB L1, 128KB L2, and 1408KB L3 cache per core. 
        We haven't further optimized for modern CPUs with larger caches, as our data must still be compatible with the older systems.

        Args:
            image_size (Tuple[int, int, int, int]): Image shape. Use a 2D, 3D,
                or 4D size; 2D/3D inputs are internally expanded.
            patch_size (Union[Tuple[int, int], Tuple[int, int, int]]): Patch
                size for spatial dimensions. Use a 2-tuple (x, y) or 3-tuple
                (x, y, z).
            bytes_per_pixel (int): Number of bytes per element. Defaults to 4
                for float32.
            l1_cache_size_per_core_in_bytes (int): L1 cache per core in bytes.
            l3_cache_size_per_core_in_bytes (int): L3 cache per core in bytes.
            safety_factor (float): Safety factor to avoid filling caches.

        Returns:
            Tuple[Tuple[int, ...], Tuple[int, ...]]: Recommended chunk size and block size.
        """

        num_squeezes = 0

        if len(image_size) == 2:
            image_size = (1, 1, *image_size)
            num_squeezes = 2

        if len(image_size) == 3:
            image_size = (1, *image_size)
            num_squeezes = 1

        if len(image_size) != 4:
            raise RuntimeError("Image size must be 4D.")
        
        if not (len(patch_size) == 2 or len(patch_size) == 3):
            raise RuntimeError("Patch size must be 2D or 3D.")

        num_channels = image_size[0]
        if len(patch_size) == 2:
            patch_size = [1, *patch_size]
        patch_size = np.array(patch_size)
        block_size = np.array((num_channels, *[2 ** (max(0, math.ceil(math.log2(i)))) for i in patch_size]))

        # shrink the block size until it fits in L1
        estimated_nbytes_block = np.prod(block_size) * bytes_per_pixel
        while estimated_nbytes_block > (l1_cache_size_per_core_in_bytes * safety_factor):
            # pick largest deviation from patch_size that is not 1
            axis_order = np.argsort(block_size[1:] / patch_size)[::-1]
            idx = 0
            picked_axis = axis_order[idx]
            while block_size[picked_axis + 1] == 1 or block_size[picked_axis + 1] == 1:
                idx += 1
                picked_axis = axis_order[idx]
            # now reduce that axis to the next lowest power of 2
            block_size[picked_axis + 1] = 2 ** (max(0, math.floor(math.log2(block_size[picked_axis + 1] - 1))))
            block_size[picked_axis + 1] = min(block_size[picked_axis + 1], image_size[picked_axis + 1])
            estimated_nbytes_block = np.prod(block_size) * bytes_per_pixel

        block_size = np.array([min(i, j) for i, j in zip(image_size, block_size)])

        # note: there is no use extending the chunk size to 3d when we have a 2d patch size! This would unnecessarily
        # load data into L3
        # now tile the blocks into chunks until we hit image_size or the l3 cache per core limit
        chunk_size = deepcopy(block_size)
        estimated_nbytes_chunk = np.prod(chunk_size) * bytes_per_pixel
        while estimated_nbytes_chunk < (l3_cache_size_per_core_in_bytes * safety_factor):
            if patch_size[0] == 1 and all([i == j for i, j in zip(chunk_size[2:], image_size[2:])]):
                break
            if all([i == j for i, j in zip(chunk_size, image_size)]):
                break
            # find axis that deviates from block_size the most
            axis_order = np.argsort(chunk_size[1:] / block_size[1:])
            idx = 0
            picked_axis = axis_order[idx]
            while chunk_size[picked_axis + 1] == image_size[picked_axis + 1] or patch_size[picked_axis] == 1:
                idx += 1
                picked_axis = axis_order[idx]
            chunk_size[picked_axis + 1] += block_size[picked_axis + 1]
            chunk_size[picked_axis + 1] = min(chunk_size[picked_axis + 1], image_size[picked_axis + 1])
            estimated_nbytes_chunk = np.prod(chunk_size) * bytes_per_pixel
            if np.mean([i / j for i, j in zip(chunk_size[1:], patch_size)]) > 1.5:
                # chunk size should not exceed patch size * 1.5 on average
                chunk_size[picked_axis + 1] -= block_size[picked_axis + 1]
                break
        # better safe than sorry
        chunk_size = [min(i, j) for i, j in zip(image_size, chunk_size)]

        block_size = block_size[num_squeezes:]
        chunk_size = chunk_size[num_squeezes:]

        return [int(value) for value in chunk_size], [int(value) for value in block_size]



"""A standardized blosc2 image reader and writer for medical images.."""

from importlib import metadata as _metadata
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from med_blosc2.med_blosc2 import MedBlosc2, MED_BLOSC2_DEFAULT_PATCH_SIZE
    from med_blosc2.meta import Meta, MetaBlosc2, MetaSpatial
    from med_blosc2.utils import is_serializable

__all__ = [
    "__version__",
    "MedBlosc2",
    "MED_BLOSC2_DEFAULT_PATCH_SIZE",
    "Meta",
    "MetaBlosc2",
    "MetaSpatial",
    "is_serializable",
]

try:
    __version__ = _metadata.version(__name__)
except _metadata.PackageNotFoundError:  # pragma: no cover - during editable installs pre-build
    __version__ = "0.0.0"


def __getattr__(name: str):
    if name in {"MedBlosc2", "MED_BLOSC2_DEFAULT_PATCH_SIZE"}:
        from med_blosc2.med_blosc2 import MedBlosc2, MED_BLOSC2_DEFAULT_PATCH_SIZE

        return MedBlosc2 if name == "MedBlosc2" else MED_BLOSC2_DEFAULT_PATCH_SIZE
    if name in {"Meta", "MetaBlosc2", "MetaSpatial"}:
        from med_blosc2.meta import Meta, MetaBlosc2, MetaSpatial

        return {"Meta": Meta, "MetaBlosc2": MetaBlosc2, "MetaSpatial": MetaSpatial}[name]
    if name == "is_serializable":
        from med_blosc2.utils import is_serializable

        return is_serializable
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


def __dir__():
    return sorted(__all__)



import json


def is_serializable(d: dict) -> bool:
    """Checks whether a dictionary is JSON-serializable.

    Args:
        d (dict): Input dictionary to test.

    Returns:
        bool: True when serializable, otherwise False.
    """
    try:
        json.dumps(d)
        return True
    except (TypeError, OverflowError):
        return False


from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Dict, List, Mapping, Optional, Union
import numpy as np
from med_blosc2.utils import is_serializable


@dataclass(slots=True)
class MetaBlosc2:
    chunk_size: Optional[list] = None
    block_size: Optional[list] = None
    patch_size: Optional[list] = None

    def __post_init__(self) -> None:
        self._validate_and_cast()

    def __repr__(self) -> str:
        return repr(self.to_dict())

    def to_dict(self, *, include_none: bool = True) -> Dict[str, Any]:
        out: Dict[str, Any] = {
            "chunk_size": self.chunk_size,
            "block_size": self.block_size,
            "patch_size": self.patch_size,
        }
        if not include_none:
            out = {k: v for k, v in out.items() if v is not None}
        return out

    def _validate_and_cast(self, ndims: Optional[int] = None) -> None:
        if self.chunk_size is not None:
            self.chunk_size = _cast_to_list(self.chunk_size, "meta._blosc2.chunk_size")
            _validate_float_int_list(self.chunk_size, f"meta._blosc2.chunk_size", ndims)
        if self.block_size is not None:
            self.block_size = _cast_to_list(self.block_size, "meta._blosc2.block_size")
            _validate_float_int_list(self.block_size, f"meta._blosc2.block_size", ndims)
        if self.patch_size is not None:
            self.patch_size = _cast_to_list(self.patch_size, "meta._blosc2.patch_size")
            _validate_float_int_list(self.patch_size, f"meta._blosc2.patch_size", ndims)

    @classmethod
    def from_dict(cls, d: Mapping[str, Any], *, strict: bool = True) -> MetaBlosc2:
        if not isinstance(d, Mapping):
            raise TypeError(f"MetaBlosc2.from_dict expects a mapping, got {type(d).__name__}")
        known = {"chunk_size", "block_size", "patch_size"}
        d = dict(d)
        unknown = set(d.keys()) - known
        if unknown and strict:
            raise KeyError(f"Unknown MetaBlosc2 keys in from_dict: {sorted(unknown)}")
        return cls(
            chunk_size=d.get("chunk_size"),
            block_size=d.get("block_size"),
            patch_size=d.get("patch_size"),
        )


@dataclass(slots=True)
class MetaSpatial:
    spacing: Optional[List] = None
    origin: Optional[List] = None
    direction: Optional[List[List]] = None

    def __post_init__(self) -> None:
        self._validate_and_cast()

    def __repr__(self) -> str:
        return repr(self.to_dict())

    def to_dict(self, *, include_none: bool = True) -> Dict[str, Any]:
        out: Dict[str, Any] = {
            "spacing": self.spacing,
            "origin": self.origin,
            "direction": self.direction,
        }
        if not include_none:
            out = {k: v for k, v in out.items() if v is not None}
        return out

    def _validate_and_cast(self, ndims: Optional[int] = None) -> None:
        if self.spacing is not None:
            self.spacing = _cast_to_list(self.spacing, "meta.spatial.spacing")
            _validate_float_int_list(self.spacing, "meta.spatial.spacing", ndims)
        if self.origin is not None:
            self.origin = _cast_to_list(self.origin, "meta.spatial.origin")
            _validate_float_int_list(self.origin, "meta.spatial.origin", ndims)
        if self.direction is not None:
            self.direction = _cast_to_list(self.direction, "meta.spatial.direction")
            _validate_float_int_matrix(self.direction, "meta.spatial.direction", ndims)

    @classmethod
    def from_dict(cls, d: Mapping[str, Any], *, strict: bool = True) -> MetaSpatial:
        if not isinstance(d, Mapping):
            raise TypeError(f"MetaSpatial.from_dict expects a mapping, got {type(d).__name__}")
        known = {"spacing", "origin", "direction"}
        d = dict(d)
        unknown = set(d.keys()) - known
        if unknown and strict:
            raise KeyError(f"Unknown MetaSpatial keys in from_dict: {sorted(unknown)}")
        return cls(
            spacing=d.get("spacing"),
            origin=d.get("origin"),
            direction=d.get("direction"),
        )


@dataclass(slots=True)
class MetaStatistics:
    min: Optional[float] = None
    max: Optional[float] = None
    mean: Optional[float] = None
    median: Optional[float] = None
    std: Optional[float] = None
    percentile_min: Optional[float] = None
    percentile_max: Optional[float] = None
    percentile_mean: Optional[float] = None
    percentile_median: Optional[float] = None
    percentile_std: Optional[float] = None

    def __post_init__(self) -> None:
        self._validate_and_cast()

    def __repr__(self) -> str:
        return repr(self.to_dict())

    def to_dict(self, *, include_none: bool = True) -> Dict[str, Any]:
        out: Dict[str, Any] = {
            "min": self.min,
            "max": self.max,
            "mean": self.mean,
            "median": self.median,
            "std": self.std,
            "percentile_min": self.percentile_min,
            "percentile_max": self.percentile_max,
            "percentile_mean": self.percentile_mean,
            "percentile_median": self.percentile_median,
            "percentile_std": self.percentile_std,
        }
        if not include_none:
            out = {k: v for k, v in out.items() if v is not None}
        return out

    def _validate_and_cast(self) -> None:
        for name in (
            "min",
            "max",
            "mean",
            "median",
            "std",
            "percentile_min",
            "percentile_max",
            "percentile_mean",
            "percentile_median",
            "percentile_std",
        ):
            value = getattr(self, name)
            if value is not None and not isinstance(value, (float, int)):
                raise TypeError(f"meta.stats.{name} must be a float or int")

    @classmethod
    def from_dict(cls, d: Mapping[str, Any], *, strict: bool = True) -> MetaStatistics:
        if not isinstance(d, Mapping):
            raise TypeError(f"MetaStatistics.from_dict expects a mapping, got {type(d).__name__}")
        known = {
            "min",
            "max",
            "mean",
            "median",
            "std",
            "percentile_min",
            "percentile_max",
            "percentile_mean",
            "percentile_median",
            "percentile_std",
        }
        d = dict(d)
        unknown = set(d.keys()) - known
        if unknown and strict:
            raise KeyError(f"Unknown MetaStatistics keys in from_dict: {sorted(unknown)}")
        return cls(
            min=d.get("min"),
            max=d.get("max"),
            mean=d.get("mean"),
            median=d.get("median"),
            std=d.get("std"),
            percentile_min=d.get("percentile_min"),
            percentile_max=d.get("percentile_max"),
            percentile_mean=d.get("percentile_mean"),
            percentile_median=d.get("percentile_median"),
            percentile_std=d.get("percentile_std"),
        )


@dataclass(slots=True)
class MetaBbox:
    bboxes: Optional[List[List[List[int]]]] = None

    def __post_init__(self) -> None:
        self._validate_and_cast()

    def __iter__(self):
        return iter(self.bboxes or [])

    def __getitem__(self, index):
        if self.bboxes is None:
            raise TypeError("meta.bbox is None")
        return self.bboxes[index]

    def __setitem__(self, index, value):
        if self.bboxes is None:
            raise TypeError("meta.bbox is None")
        self.bboxes[index] = value

    def __len__(self):
        return len(self.bboxes or [])

    def __repr__(self) -> str:
        return repr(self.bboxes)

    def to_list(self) -> Optional[List[List[List[int]]]]:
        return self.bboxes

    def _validate_and_cast(self, ndims: Optional[int] = None) -> None:
        if self.bboxes is None:
            return
        self.bboxes = _cast_to_list(self.bboxes, "meta.bbox.bboxes")
        if not isinstance(self.bboxes, list):
            raise TypeError("meta.bbox.bboxes must be a list of bboxes")
        for bbox in self.bboxes:
            if not isinstance(bbox, list):
                raise TypeError("meta.bbox.bboxes must be a list of bboxes")
            if ndims is not None and len(bbox) != ndims:
                raise ValueError(f"meta.bbox.bboxes entries must have length {ndims}")
            for row in bbox:
                if not isinstance(row, list):
                    raise TypeError("meta.bbox.bboxes must be a list of lists")
                if len(row) != 2:
                    raise ValueError("meta.bbox.bboxes entries must have length 2 per dimension")
                for item in row:
                        if isinstance(item, bool) or not isinstance(item, int):
                            raise TypeError("meta.bbox.bboxes must contain only ints")

    @classmethod
    def from_list(cls, bboxes: Any) -> MetaBbox:
        return cls(bboxes=bboxes)


@dataclass(slots=True)
class Meta:
    image: Optional[Dict[str, Any]] = None
    spatial: MetaSpatial = field(default_factory=MetaSpatial)
    stats: Optional[Union[dict, MetaStatistics]] = None
    bbox: Optional[MetaBbox] = None
    is_seg: Optional[bool] = None
    _blosc2: MetaBlosc2 = field(default_factory=MetaBlosc2)
    _med_blosc2_version: Optional[str] = None

    # controlled escape hatch for future/experimental metadata
    extra: Dict[str, Any] = field(default_factory=dict)

    def __post_init__(self) -> None:
        # Validate anything passed in the constructor
        for name in ("image",):
            val = getattr(self, name)
            if val is not None:
                if not isinstance(val, dict):
                    raise TypeError(f"meta.{name} must be a dict or None, got {type(val).__name__}")
                if not is_serializable(val):
                    raise TypeError(f"meta.{name} is not JSON-serializable")
        if self.stats is not None:
            if isinstance(self.stats, MetaStatistics):
                pass
            elif isinstance(self.stats, Mapping):
                self.stats = MetaStatistics.from_dict(self.stats, strict=False)
            else:
                raise TypeError(f"meta.stats must be a MetaStatistics or mapping, got {type(self.stats).__name__}")
        if self.bbox is not None:
            if isinstance(self.bbox, MetaBbox):
                pass
            elif isinstance(self.bbox, (list, tuple)) or (np is not None and isinstance(self.bbox, np.ndarray)):
                self.bbox = MetaBbox(bboxes=self.bbox)
            else:
                raise TypeError(f"meta.bbox must be a MetaBbox or list-like, got {type(self.bbox).__name__}")

        if self.spatial is None:
            self.spatial = MetaSpatial()
        if not isinstance(self.spatial, MetaSpatial):
            raise TypeError(f"meta.spatial must be a MetaSpatial, got {type(self.spatial).__name__}")

        if self._blosc2 is None:
            self._blosc2 = MetaBlosc2()
        if not isinstance(self._blosc2, MetaBlosc2):
            raise TypeError(f"meta._blosc2 must be a MetaBlosc2, got {type(self._blosc2).__name__}")

        if self.is_seg is not None and not isinstance(self.is_seg, bool):
            raise TypeError("meta.is_seg must be a bool or None")
        if self._med_blosc2_version is not None and not isinstance(self._med_blosc2_version, str):
            raise TypeError("meta._med_blosc2_version must be a str or None")

        if not isinstance(self.extra, dict):
            raise TypeError(f"meta.extra must be a dict, got {type(self.extra).__name__}")
        if not is_serializable(self.extra):
            raise TypeError("meta.extra is not JSON-serializable")

    def __repr__(self) -> str:
        return repr(self.to_dict())

    def set(self, key: str, value: Any) -> None:
        """
        Set a known meta section explicitly (typos raise).
        Ensures the provided value is JSON-serializable.
        """
        if not hasattr(self, key) and key not in {"_blosc2", "_med_blosc2_version"}:
            raise AttributeError(f"Unknown meta section: {key!r}")
        if key == "extra":
            raise AttributeError("Use meta.extra[...] to add to extra")
        if key == "spatial":
            if isinstance(value, MetaSpatial):
                setattr(self, key, value)
                return
            if isinstance(value, Mapping):
                setattr(self, key, MetaSpatial.from_dict(value, strict=False))
                return
            raise TypeError("meta.spatial must be a MetaSpatial or mapping")
        if key == "stats":
            if isinstance(value, MetaStatistics):
                self.stats = value
                return
            if isinstance(value, Mapping):
                self.stats = MetaStatistics.from_dict(value, strict=False)
                return
            raise TypeError("meta.stats must be a MetaStatistics or mapping")
        if key == "bbox":
            if isinstance(value, MetaBbox):
                self.bbox = value
                return
            if isinstance(value, (list, tuple)) or (np is not None and isinstance(value, np.ndarray)):
                self.bbox = MetaBbox(bboxes=value)
                return
            raise TypeError("meta.bbox must be a MetaBbox or list-like")
        if key == "_blosc2":
            if isinstance(value, MetaBlosc2):
                self._blosc2 = value
                return
            if isinstance(value, Mapping):
                self._blosc2 = MetaBlosc2.from_dict(value, strict=False)
                return
            raise TypeError("meta._blosc2 must be a MetaBlosc2 or mapping")
        if key == "is_seg":
            if value is not None and not isinstance(value, bool):
                raise TypeError("meta.is_seg must be a bool or None")
            setattr(self, key, value)
            return
        if key == "_med_blosc2_version":
            if value is not None and not isinstance(value, str):
                raise TypeError("meta._med_blosc2_version must be a str or None")
            self._med_blosc2_version = value
            return

        value_dict = dict(value)

        if not is_serializable(value_dict):
            raise TypeError(f"meta.{key} is not JSON-serializable")

        setattr(self, key, value_dict)

    def to_dict(self, *, include_none: bool = True) -> Dict[str, Any]:
        """
        Convert to a plain dict. All entries are guaranteed JSON-serializable
        due to validation in __post_init__ and set().
        """
        out: Dict[str, Any] = {
            "image": self.image,
            "stats": self.stats.to_dict() if self.stats is not None else None,
            "bbox": self.bbox.to_list() if self.bbox is not None else None,
            "is_seg": self.is_seg,
            "spatial": self.spatial.to_dict(),
            "_blosc2": self._blosc2.to_dict(),
            "_med_blosc2_version": self._med_blosc2_version,
            "extra": self.extra,
        }

        if not include_none:
            out = {k: v for k, v in out.items() if v is not None and not (k == "extra" and v == {})}

        return out

    def _validate_and_cast(self, ndims: int) -> None:
        self.spatial._validate_and_cast(ndims)
        if self.bbox is not None:
            self.bbox._validate_and_cast(ndims)
        self._blosc2._validate_and_cast(ndims)

    @classmethod
    def from_dict(cls, d: Mapping[str, Any], *, strict: bool = True) -> Meta:
        """
        Construct Meta from a dict.

        Args:
            d: Mapping with keys in {"image", "stats", "bbox", "spatial",
                "_blosc2", "_med_blosc2_version", "is_seg", "extra"}.
            strict: If True, unknown keys raise. If False, unknown keys go into extra.

        Returns:
            Meta instance (validated).
        """
        if not isinstance(d, Mapping):
            raise TypeError(f"from_dict expects a mapping, got {type(d).__name__}")

        known = {"image", "stats", "bbox", "spatial", "_blosc2", "_med_blosc2_version", "is_seg", "extra"}
        d = dict(d)
        unknown = set(d.keys()) - known

        if unknown and strict:
            raise KeyError(f"Unknown meta keys in from_dict: {sorted(unknown)}")

        extra = dict(d.get("extra") or {})
        if unknown and not strict:
            for k in unknown:
                extra[k] = d[k]

        spatial = d.get("spatial")
        if spatial is None:
            spatial = MetaSpatial()
        else:
            spatial = MetaSpatial.from_dict(spatial, strict=strict)

        stats = d.get("stats")
        if stats is None:
            stats = None
        else:
            stats = MetaStatistics.from_dict(stats, strict=strict)

        bbox = d.get("bbox")
        if bbox is None:
            bbox = None
        else:
            bbox = MetaBbox.from_list(bbox)

        _blosc2 = d.get("_blosc2")
        if _blosc2 is None:
            _blosc2 = MetaBlosc2()
        else:
            _blosc2 = MetaBlosc2.from_dict(_blosc2, strict=strict)

        return cls(
            image=d.get("image"),
            stats=stats,
            bbox=bbox,
            is_seg=d.get("is_seg"),
            spatial=spatial,
            _blosc2=_blosc2,
            _med_blosc2_version=d.get("_med_blosc2_version"),            
            extra=extra,
        )

    def copy_from(self, other: Meta) -> None:
        """
        Copy fields from another Meta if they are not set on this instance.
        """
        if self.image is None:
            self.image = other.image
        if self.stats is None:
            self.stats = other.stats
        if self.bbox is None:
            self.bbox = other.bbox
        if self.is_seg is None:
            self.is_seg = other.is_seg
        if self.spatial is None:
            self.spatial = other.spatial
        elif other.spatial is not None:
            if self.spatial.spacing is None:
                self.spatial.spacing = other.spatial.spacing
            if self.spatial.origin is None:
                self.spatial.origin = other.spatial.origin
            if self.spatial.direction is None:
                self.spatial.direction = other.spatial.direction
        if self._blosc2 is None:
            self._blosc2 = other._blosc2
        elif other._blosc2 is not None:
            if self._blosc2.chunk_size is None:
                self._blosc2.chunk_size = other._blosc2.chunk_size
            if self._blosc2.block_size is None:
                self._blosc2.block_size = other._blosc2.block_size
            if self._blosc2.patch_size is None:
                self._blosc2.patch_size = other._blosc2.patch_size
        if self._med_blosc2_version is None:
            self._med_blosc2_version = other._med_blosc2_version
        if self.extra == {}:
            self.extra = other.extra


def _cast_to_list(value: Any, label: str):
    if isinstance(value, list):
        out = value
    elif isinstance(value, tuple):
        out = list(value)
    elif np is not None and isinstance(value, np.ndarray):
        out = value.tolist()
    else:
        raise TypeError(f"{label} must be a list, tuple, or numpy array")

    if not isinstance(out, list):
        raise TypeError(f"{label} must be a list, tuple, or numpy array")

    for idx, item in enumerate(out):
        if isinstance(item, (list, tuple)) or (np is not None and isinstance(item, np.ndarray)):
            out[idx] = _cast_to_list(item, label)
    return out


def _validate_float_int_list(value: Any, label: str, ndims: Optional[int] = None) -> None:
    if not isinstance(value, list):
        raise TypeError(f"{label} must be a list")
    if ndims is not None and len(value) != ndims:
        raise ValueError(f"{label} must have length {ndims}")
    for item in value:
        if not isinstance(item, (float, int)):
            raise TypeError(f"{label} must contain only floats or ints")


def _validate_float_int_matrix(value: Any, label: str, ndims: Optional[int] = None) -> None:
    if not isinstance(value, list):
        raise TypeError(f"{label} must be a list of lists")
    if ndims is not None and len(value) != ndims:
        raise ValueError(f"{label} must have shape [{ndims}, {ndims}]")
    for row in value:
        if not isinstance(row, list):
            raise TypeError(f"{label} must be a list of lists")
        if ndims is not None and len(row) != ndims:
            raise ValueError(f"{label} must have shape [{ndims}, {ndims}]")
        for item in row:
            if not isinstance(item, (float, int)):
                raise TypeError(f"{label} must contain only floats or ints")



import tempfile
import unittest
from pathlib import Path

import numpy as np

from med_blosc2 import MedBlosc2, MED_BLOSC2_DEFAULT_PATCH_SIZE
from med_blosc2.meta import Meta


def _make_array(shape=(16, 32, 32), seed=0):
    rng = np.random.default_rng(seed)
    return rng.random(shape, dtype=np.float32)


class TestUsage(unittest.TestCase):
    def test_default_usage_save_load(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            array = _make_array()
            image = MedBlosc2(array)

            path = Path(tmpdir) / "sample.mb2nd"
            image.save(path)

            loaded = MedBlosc2(path)
            self.assertEqual(loaded.shape, array.shape)

    def test_mmap_loading(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            array = _make_array()
            path = Path(tmpdir) / "sample.mb2nd"
            MedBlosc2(array).save(path)

            loaded = MedBlosc2(path, mmap=True)
            self.assertFalse(isinstance(loaded.array, np.ndarray))

    def test_loading_and_saving(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            array = _make_array()
            src = Path(tmpdir) / "sample.mb2nd"
            dst = Path(tmpdir) / "copy.mb2nd"

            MedBlosc2(array).save(src)
            MedBlosc2(src).save(dst)

            self.assertTrue(dst.exists())

    def test_metadata_inspection_and_manipulation(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            array = _make_array()
            image = MedBlosc2(
                array,
                spacing=(1.0, 1.0, 1.5),
                origin=(10.0, 10.0, 30.0),
                direction=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
                meta=Meta(image={"patient_id": "123", "modality": "CT"}, is_seg=False),
            )
            image.spacing[1] = 5.3
            image.meta.image["study_id"] = "study-001"

            path = Path(tmpdir) / "with-metadata.mb2nd"
            image.save(path)

            loaded = MedBlosc2(path)
            self.assertEqual(loaded.spacing, [1.0, 5.3, 1.5])
            self.assertEqual(loaded.origin, [10.0, 10.0, 30.0])
            self.assertEqual(loaded.meta.image["study_id"], "study-001")

    def test_copy_metadata_with_override(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            array = _make_array()
            base = MedBlosc2(
                array,
                spacing=(1.0, 1.0, 1.0),
                origin=(1.0, 2.0, 3.0),
                direction=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
                meta=Meta(image={"source": "base"}, is_seg=True),
            )
            base_path = Path(tmpdir) / "base.mb2nd"
            base.save(base_path)

            base_loaded = MedBlosc2(base_path)
            array2 = _make_array(seed=1)
            image = MedBlosc2(array2, spacing=(0.8, 0.8, 1.0), copy=base_loaded)

            self.assertEqual(image.spacing, [0.8, 0.8, 1.0])
            self.assertEqual(image.origin, base_loaded.origin)
            self.assertEqual(image.direction, base_loaded.direction)
            self.assertEqual(image.meta.is_seg, base_loaded.meta.is_seg)
            self.assertEqual(image.meta.image, base_loaded.meta.image)

    def test_patch_size_default(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            array = _make_array()
            path = Path(tmpdir) / "default-patch.mb2nd"
            MedBlosc2(array).save(path)

            loaded = MedBlosc2(path)
            self.assertEqual(
                loaded.meta._blosc2.patch_size,
                [MED_BLOSC2_DEFAULT_PATCH_SIZE,] * 3,
            )

    def test_patch_size_isotropic(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            array = _make_array()
            path = Path(tmpdir) / "patch-64.mb2nd"
            MedBlosc2(array).save(path, patch_size=64)

            loaded = MedBlosc2(path)
            self.assertEqual(loaded.meta._blosc2.patch_size, [64, 64, 64])

    def test_patch_size_non_isotropic(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            array = _make_array()
            path = Path(tmpdir) / "patch-non-iso.mb2nd"
            MedBlosc2(array).save(path, patch_size=[16, 24, 32])

            loaded = MedBlosc2(path)
            self.assertEqual(loaded.meta._blosc2.patch_size, [16, 24, 32])

    def test_manual_chunk_and_block(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            array = _make_array()
            path = Path(tmpdir) / "manual-chunk-block.mb2nd"
            MedBlosc2(array).save(
                path,
                patch_size=None,
                chunk_size=(1, 16, 16),
                block_size=(1, 8, 8),
            )

            loaded = MedBlosc2(path)
            self.assertEqual(loaded.meta._blosc2.chunk_size, [1, 16, 16])
            self.assertEqual(loaded.meta._blosc2.block_size, [1, 8, 8])

    def test_b2nd_metadata_ignored_on_load(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            array = _make_array()
            path = Path(tmpdir) / "plain.b2nd"
            image = MedBlosc2(
                array,
                spacing=(1.0, 2.0, 3.0),
                origin=(1.0, 2.0, 3.0),
                direction=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
                meta=Meta(image={"tag": "value"}, is_seg=True),
            )
            image.save(path)

            loaded = MedBlosc2(path)
            self.assertIsNone(loaded.spacing)
            self.assertIsNone(loaded.origin)
            self.assertIsNone(loaded.direction)


if __name__ == "__main__":
    unittest.main()
